@@ -1,12 +1,13 @@
 #  GTK+ based frontend to software-properties
 #
-#  Copyright (c) 2004-2018 Canonical Ltd.
+#  Copyright (c) 2004-2020 Canonical Ltd.
 #                2004-2005 Michiel Sikkes
 #
 #  Author: Michiel Sikkes <michiel@eyesopened.nl>
 #          Michael Vogt <mvo@debian.org>
 #          Sebastian Heinlein <glatzor@ubuntu.com>
 #          Andrea Azzarone <andrea.azzarone@canonical.com>
+#          Olivier Tilloy <olivier.tilloy@canonical.com>
 #
 #  This program is free software; you can redistribute it and/or
 #  modify it under the terms of the GNU General Public License as
@@ -27,6 +28,7 @@ from __future__ import absolute_import, print_function
 
 import apt
 import apt_pkg
+import datetime
 import dbus
 import dbus.glib
 import dbus.mainloop.glib
@@ -37,6 +39,7 @@ import subprocess
 import logging
 import threading
 import sys
+import time
 
 import gi
 gi.require_version("Gdk", "3.0")
@@ -58,6 +61,13 @@ import softwareproperties.distro
 from softwareproperties.SoftwareProperties import SoftwareProperties
 import softwareproperties.SoftwareProperties
 
+from softwareproperties.gtk.utils import (
+    get_ua_status,
+    get_ua_service_status,
+    current_distro,
+    is_current_distro_lts,
+)
+
 from UbuntuDrivers import detect
 dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
 
@@ -83,6 +93,52 @@ RESPONSE_ADD = 2
     STORE_VISIBLE
 ) = list(range(5))
 
+# rows for the updates subscription combobox
+(
+    UPDATES_ALL,
+    UPDATES_RECOMMENDED,
+    UPDATES_SECURITY,
+    UPDATES_CUSTOM,
+) = range(4)
+
+# names of the pockets
+POCKET_SECURITY = "security"
+POCKET_UPDATES = "updates"
+POCKET_BACKPORTS = "backports"
+
+
+class UpdateSubscriptions(object):
+    security = False
+    updates = False
+    backports = False
+    other = False
+    inconsistent = False
+
+    def get_status(self):
+        if self.other:
+            return UPDATES_CUSTOM
+        if self.security and self.updates and self.backports:
+            return UPDATES_ALL
+        elif self.security and self.updates and not self.backports:
+            return UPDATES_RECOMMENDED
+        elif self.security and not self.updates and not self.backports:
+            return UPDATES_SECURITY
+        return UPDATES_CUSTOM
+
+
+def maybe_log_authentication_canceled_error(dbus_exception):
+    """Log an error if the DBus exception was raised because the user canceled
+    the authentication prompt shown when changing apt sources/configuration.
+
+    Return True if it is, so that calling code can do further processing,
+    False otherwise.
+    """
+    if dbus_exception._dbus_error_name == \
+      'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
+        logging.error("Authentication canceled, changes have not been saved")
+        return True
+    return False
+
 
 def error(parent_window, summary, msg):
     """ show a error dialog """
@@ -97,18 +153,6 @@ def error(parent_window, summary, msg):
     return False
 
 
-def get_dependencies(apt_cache, package_name, pattern=None):
-    """ Get the package dependencies, which can be filtered out by a pattern """
-    dependencies = []
-    for or_group in apt_cache[package_name].candidate.dependencies:
-        for dep in or_group:
-            if dep.rawtype in ["Depends", "PreDepends"]:
-                dependencies.append(dep.name)
-    if pattern:
-        dependencies = [ x for x in dependencies if x.find(pattern) != -1 ]
-    return dependencies
-
-
 class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
 
     def __init__(self, datadir=None, options=None, file=None, parent=None):
@@ -156,6 +200,9 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             self.initial_auto_launch = 0
             self.combobox_other_updates.set_sensitive(False)
 
+        self.sp_settings = Gio.Settings.new('com.ubuntu.SoftwareProperties')
+        self.info_bar_ubuntu_pro.set_visible(self.sp_settings.get_boolean("ubuntu-pro-banner-visible"))
+
         # get the dbus backend
         bus = dbus.SystemBus()
         proxy = bus.get_object("com.ubuntu.SoftwareProperties", "/")
@@ -347,29 +394,79 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             self.vbox_dist_comps.add(checkbox)
             checkbox.show()
 
-        # Setup the checkbuttons for the child repos / updates
-        for checkbutton in self.vbox_updates.get_children():
-            self.vbox_updates.remove(checkbutton)
+        # Setup the checkbuttons for the child repos
         for checkbutton in self.dev_box.get_children():
             self.dev_box.remove(checkbutton)
-        if len(self.distro.source_template.children) < 1:
-            self.frame_children.hide()
         for template in self.distro.source_template.children:
             # Do not show source entries in there
             if template.type == "deb-src":
                 continue
-
-            checkbox = Gtk.CheckButton(label="%s (%s)" % (template.description,
-                                                          template.name))
-            checkbox.template = template
-            self.handlers[checkbox] = checkbox.connect("toggled",
-                                                   self.on_checkbutton_child_toggled,
-                                                   template)
             if "proposed" in template.name:
+                checkbox = Gtk.CheckButton(label="%s (%s)" % (template.description,
+                                                              template.name))
+                checkbox.template = template
+                self.handlers[checkbox] = \
+                    checkbox.connect("toggled",
+                                     self.on_checkbutton_child_toggled,
+                                     template)
                 self.dev_box.add(checkbox)
-            else:
-                self.vbox_updates.add(checkbox)
-            checkbox.show()
+                checkbox.show()
+
+        status = get_ua_status()
+        if not is_current_distro_lts():
+                esm_available = False
+                esm_enabled = False
+        else:
+                (infra_available, infra_status) = get_ua_service_status("esm-infra", status=status)
+                (apps_available, apps_status) = get_ua_service_status("esm-apps", status=status)
+                esm_available = bool(infra_available or apps_available)
+                esm_enabled = "enabled" in (infra_status, apps_status)
+        distro = current_distro()
+        if esm_enabled:
+                eol_text = _("Extended Security Maintenance")
+                # EOL date should probably be UA contract expiry.
+                # This is probably sooner than ESM EOL for the distro and
+                # gives software properties dialogs a chance to interact about
+                # renewals if needed.
+                try:
+                        # Ignore timezone to simplify formatting python < 3.7
+                        # 3.7 has datetime.fromisoformat()
+                        dt, _sep, _tz = status.get("expires", "").partition("+")
+                        eol_date = datetime.datetime.strptime(
+                            dt, "%Y-%m-%dT%H:%M:%S"
+                        ).date()
+                except ValueError:
+                        print("Unable to determine UA contract expiry")
+                        eol_date = distro.eol
+        else:
+                eol_text = _("Basic Security Maintenance")
+                eol_date = distro.eol
+        self.label_esm_status.set_markup(eol_text)
+        esm_url = "https://ubuntu.com/esm"  # Non-EOL LTS generic ESM
+        today = datetime.datetime.now().date()
+        if today >= eol_date:
+                if esm_available:
+                        # EOL LTS uses release-specific ESM ubuntu.com/XX-YY
+                        distro_ver = distro.version.replace(' LTS', '')
+                        esm_url = "https://ubuntu.com/%s" % distro_ver.replace(".", "-")
+                eol_expiry_text = _("Ended %s - extend or upgrade now") % eol_date.strftime("%x")
+        elif today >= eol_date - datetime.timedelta(days=60):
+                eol_expiry_text = _("Ends %s - extend or upgrade soon") % eol_date.strftime("%x")
+        else:
+                eol_expiry_text = _("Active until %s") % eol_date.strftime("%x")
+        self.label_eol.set_label(eol_expiry_text)
+        self.label_esm_subscribe.set_markup(
+                "<a href=\"%s\">%s</a>" % (esm_url, _("Extend…"))
+        )
+        self.label_esm_subscribe.set_visible(
+                esm_available and not esm_enabled
+        )
+        eol_expiry_text = _("Ended %s") % eol_date.strftime("%x")
+
+        # Setup the combo box for updates subscriptions
+        combobox = self.combobox_updates_subscription
+        self.handlers[combobox] = combobox.connect(
+            "changed", self.on_combobox_updates_subscription_changed)
 
         # setup the server chooser
         cell = Gtk.CellRendererText()
@@ -397,12 +494,6 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         """Fill the distro user interface with life"""
         self.block_handlers()
 
-        # Enable or disable the child source checkbuttons
-        for checkbox in self.vbox_updates.get_children():
-            (active, inconsistent) = self.get_comp_child_state(checkbox.template)
-            checkbox.set_active(active)
-            checkbox.set_inconsistent(inconsistent)
-
         # Enable or disable the child source checkbuttons
         for checkbox in self.dev_box.get_children():
             (active, inconsistent) = self.get_comp_child_state(checkbox.template)
@@ -416,14 +507,48 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             checkbox.set_inconsistent(inconsistent)
             checkbox.set_active(active)
 
+        # Refresh the combo box for updates subscriptions
+        subscriptions = UpdateSubscriptions()
+        for template in self.distro.source_template.children:
+            # Do not show source entries in there
+            if template.type == "deb-src":
+                continue
+            pocket = template.name.split('-')[-1]
+            (active, inconsistent) = self.get_comp_child_state(template)
+            subscriptions.inconsistent |= inconsistent
+            if active:
+                if pocket == POCKET_SECURITY:
+                    subscriptions.security = True
+                elif pocket == POCKET_UPDATES:
+                    subscriptions.updates = True
+                elif pocket == POCKET_BACKPORTS:
+                    subscriptions.backports = True
+                else:
+                    subscriptions.other = True
+
+        status = subscriptions.get_status()
+        combobox = self.combobox_updates_subscription
+        model = combobox.get_model()
+        if status == UPDATES_CUSTOM:
+            try:
+                model.get_iter_from_string(str(UPDATES_CUSTOM))
+            except ValueError:
+                model.append([_("Custom")])
+        else:
+            try:
+                model.remove(model.get_iter_from_string(str(UPDATES_CUSTOM)))
+            except ValueError:
+                pass
+        combobox.set_active(status)
+        combobox.set_sensitive(not subscriptions.inconsistent)
+
         # If no components are enabled there will be no need for updates
         # and source code
         if len(self.distro.enabled_comps) < 1:
-            self.vbox_updates.set_sensitive(False)
+            self.combobox_updates_subscription.set_sensitive(False)
             self.dev_box.set_sensitive(False)
             self.checkbutton_source_code.set_sensitive(False)
         else:
-            self.vbox_updates.set_sensitive(True)
             self.dev_box.set_sensitive(True)
             self.checkbutton_source_code.set_sensitive(True)
 
@@ -510,9 +635,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             if state != self.get_update_automation_level():
                 self.backend.SetUpdateAutomationLevel(state)
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
-
+            if maybe_log_authentication_canceled_error(e):
                 combo_handler = self.handlers[self.combobox_security_updates]
                 self.combobox_security_updates.handler_block(combo_handler)
                 self.set_security_update_level()
@@ -537,15 +660,52 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         try:
             self.backend.SetReleaseUpgradesPolicy(i)
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
-
+            if maybe_log_authentication_canceled_error(e):
                 combo_handler = self.handlers[self.combobox_release_upgrades]
                 self.combobox_release_upgrades.handler_block(combo_handler)
                 i = self.get_release_upgrades_policy()
                 self.combobox_release_upgrades.set_active(i)
                 self.combobox_release_upgrades.handler_unblock(combo_handler)
 
+    def on_combobox_updates_subscription_changed(self, combobox):
+        index = combobox.get_active()
+        if index != UPDATES_CUSTOM:
+            model = combobox.get_model()
+            try:
+                model.remove(model.get_iter_from_string(str(UPDATES_CUSTOM)))
+            except ValueError:
+                pass
+
+        def _toggle_child_source(pocket, enabled):
+            source = "{}-{}".format(self.distro.codename, pocket)
+            if enabled:
+                logging.info("enabling {}".format(source))
+                self.backend.EnableChildSource(source)
+            else:
+                logging.info("disabling {}".format(source))
+                self.backend.DisableChildSource(source)
+
+        try:
+            if index == UPDATES_ALL:
+                _toggle_child_source(POCKET_SECURITY, True)
+                _toggle_child_source(POCKET_UPDATES, True)
+                _toggle_child_source(POCKET_BACKPORTS, True)
+            elif index == UPDATES_RECOMMENDED:
+                _toggle_child_source(POCKET_SECURITY, True)
+                _toggle_child_source(POCKET_UPDATES, True)
+                _toggle_child_source(POCKET_BACKPORTS, False)
+            elif index == UPDATES_SECURITY:
+                _toggle_child_source(POCKET_SECURITY, True)
+                _toggle_child_source(POCKET_UPDATES, False)
+                _toggle_child_source(POCKET_BACKPORTS, False)
+            for template in self.distro.source_template.children:
+                if template.type == "deb-src":
+                    continue
+                pocket = template.name.split('-')[-1]
+                if pocket not in (POCKET_SECURITY, POCKET_UPDATES, POCKET_BACKPORTS):
+                    _toggle_child_source(pocket, False)
+        except dbus.DBusException as e:
+            maybe_log_authentication_canceled_error(e)
 
     def on_combobox_server_changed(self, combobox):
         """
@@ -573,8 +733,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                 self.active_server = combobox.get_active()
                 self.backend.ChangeMainDownloadServer(uri)
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
         # mvo: is this still needed?
         #else:
@@ -591,8 +750,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             else:
                 self.backend.DisableComponent(comp)
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def on_checkbutton_child_toggled(self, checkbutton, template):
         """
@@ -604,8 +762,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             else:
                 self.backend.EnableChildSource(template.name)
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def on_checkbutton_source_code_toggled(self, checkbutton):
         """ Disable or enable the source code for all sources """
@@ -615,8 +772,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             else:
                 self.backend.DisableSourceCodeSources()
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def on_checkbutton_popcon_toggled(self, widget):
         """ The user clicked on the popcon paritipcation button """
@@ -626,8 +782,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             try:
                 self.backend.SetPopconPariticipation(widget.get_active())
             except dbus.DBusException as e:
-                if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                    logging.error("Authentication canceled, changes have not been saved")
+                maybe_log_authentication_canceled_error(e)
 
     def open_file(self, file):
         """Show a confirmation for adding the channels of the specified file"""
@@ -645,8 +800,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                     for source in new_sources:
                         self.backend.AddSourceFromLine(str(source))
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def on_sources_drag_data_received(self, widget, context, x, y,
                                       selection, target_type, timestamp):
@@ -744,14 +898,20 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
 
     def on_isv_source_toggled(self, cell_toggle, path, store):
         """Enable or disable the selected channel"""
+
+        #FIXME Gtk is doing something that takes a lock while processing
+        # the events, the polkit auth dialog can't be displayed in return,
+        # we don't have a proper fix and it might require GTK changes,
+        # meanwhile we want a working interface so workaround with a sleep
+        # https://launchpad.net/bugs/1727908
+        time.sleep(0.3)
         #FIXME cdroms need to disable the comps in the childs and sources
         iter = store.get_iter((int(path),))
         source_entry = store.get_value(iter, STORE_SOURCE)
         try:
             self.backend.ToggleSourceUse(str(source_entry))
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def init_keys(self):
         """Setup the user interface parts needed for the key handling"""
@@ -812,8 +972,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         try:
             self.backend.Revert()
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
         if self.settings:
             self.settings.set_int("regular-auto-launch-interval", self.initial_auto_launch)
         self.show_auto_update_level()
@@ -897,9 +1056,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             try:
                 self.backend.SetUpdateInterval(value)
             except dbus.DBusException as e:
-                if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                    logging.error("Authentication canceled, changes have not been saved")
-
+                if maybe_log_authentication_canceled_error(e):
                     update_days = self.get_update_interval()
                     combo_handler = self.handlers[self.combobox_update_interval]
                     for key in self.combobox_interval_mapping:
@@ -918,8 +1075,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             try:
                 self.backend.AddSourceFromLine(line)
             except dbus.DBusException as e:
-                if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                    logging.error("Authentication canceled, changes have not been saved")
+                maybe_log_authentication_canceled_error(e)
 
     def on_edit_clicked(self, widget):
         """Show a dialog to edit an ISV source"""
@@ -935,8 +1091,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                 self.backend.ReplaceSourceEntry(str(old_source_entry),
                                                 str(dialog.new_source_entry))
             except dbus.DBusException as e:
-                if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                    logging.error("Authentication canceled, changes have not been saved")
+                maybe_log_authentication_canceled_error(e)
 
     # FIXME:outstanding from merge
     def on_isv_source_activated(self, treeview, path, column):
@@ -976,8 +1131,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
             try:
                 self.backend.RemoveSource(str(source_entry))
             except dbus.DBusException as e:
-                if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                    logging.error("Authentication canceled, changes have not been saved")
+                maybe_log_authentication_canceled_error(e)
 
     def add_key_clicked(self, widget):
         """Provide a file chooser that allows to add the gnupg of a trusted
@@ -1000,8 +1154,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                           _("The selected file may not be a GPG key file "
                             "or it might be corrupt."))
             except dbus.DBusException as e:
-                if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                    logging.error("Authentication canceled, changes have not been saved")
+                maybe_log_authentication_canceled_error(e)
 
     def remove_key_clicked(self, widget):
         """Remove a trusted software vendor key"""
@@ -1017,16 +1170,14 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                       _("The key you selected could not be removed. "
                         "Please report this as a bug."))
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def on_restore_clicked(self, widget):
         """Restore the original keys"""
         try:
             self.backend.UpdateKeys()
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def on_delete_event(self, widget, args):
         """Close the window if requested"""
@@ -1054,8 +1205,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
         try:
             self.backend.AddCdromSource()
         except dbus.DBusException as e:
-            if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
-                logging.error("Authentication canceled, changes have not been saved")
+            maybe_log_authentication_canceled_error(e)
 
     def on_driver_changes_progress(self, progress, ptype, data=None):
         #print(progress)
@@ -1101,7 +1251,7 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
                 # We need to collect its dependencies, so that
                 # we can uninstall the driver properly.
                 if 'nvidia' in pkg.shortname:
-                    for dep in get_dependencies(self.apt_cache, pkg.shortname, 'nvidia'):
+                    for dep in self.get_dependencies(self.apt_cache, pkg.shortname, 'nvidia'):
                         dep_pkg = self.apt_cache[dep]
                         if dep_pkg.is_installed:
                             removals.append(self.get_package_id(dep_pkg.installed))
@@ -1496,3 +1646,6 @@ class SoftwarePropertiesGtk(SoftwareProperties, SimpleGtkbuilderApp):
     def init_livepatch(self):
         self.livepatch_page = LivepatchPage(self)
 
+    def on_info_bar_ubuntu_pro_response(self, info_bar, response_id):
+        self.sp_settings.set_boolean("ubuntu-pro-banner-visible", False)
+        info_bar.hide()
